/* Copyright (C) 2005-2006  Emiliano Betti and Marco Cesati
   Copyright (C) 2003-2004  Marco Cesati
   Copyright (C) 2002-2003  Shell Technologies s.r.l.

   Authors:
            Emiliano Betti <emilianobetti@gmail.com>
            Marco Cesati   <marcocesati@gmail.com>

   Driver for the Galil 1800 PCI motion controllers (http://www.galilmc.com/)

   This is a driver for the Linux kernel 2.6. It is based on an older driver
   for the kernel 2.4, which in turn was loosely inspired by Galil's own
   Linux driver, as well as by a bunch of PCI drivers in the kernel source code

   This program is free software; you can redistribute  it and/or modify it
   under  the terms of  the GNU General  Public License as published by the
   Free Software Foundation.

   Programmers list:
                EB: Emiliano Betti
                MC: Marco Cesati

   Revision history:

    02 December 2003 - ver. 0.1.0 (MC): started new driver for kernel 2.6,
                                         write device initialization
    28 July     2005 - ver. 0.1.2 (EB): integretion with Betti's work
					(kernel 2.6.11),
					changed header files' name
    06 August   2005 - ver. 0.1.3 (EB): add debug code, add some macro,
					add "TODO" note for modifing
    08 August   2005 - ver. 0.1.4 (EB): rewriting of galil1800_card_open and
					galil1800_card_release, insert "TODO"
					note for future version
    06 October  2005 - ver. 0.1.5 (EB): debugging galil1800_card_open and
					galil1800_card_release. Manage
					(no)seeking. Preparing for ioctl,
					poll and reorganize of IO.
    18 October  2005 - ver. 0.2.0 (EB): Linux Kernel 2.6.13, re-implement of
					reset command, implemented ioctl_card,
					debug in I/O procedure, improvement of
					struct for multi-device support, first
					step for IRQ management, implemented
					event_open and event_release.
					Re-implemented all steps of init of
					the controller.
    25 October  2005 - ver. 0.2.2 (EB): Add support to special event report
					(IRQ, write buffer control, etc...).
					Debug and add on in open and release
					function of event device.
    28 October  2005 - ver. 0.2.3 (EB): Change the whole NEW event report system
					for obtaining better performance.
    02 November 2005 - ver. 0.2.5 (EB): Implemented new interrupt handlers
					(top and bottom half). Event device
					read function now is blocking.
					Implemented event_poll function.
    09 November 2005 - ver. 0.3.0 (EB): Add support for kernel buffer for 'card'
					write function. Add support for a
					"pause" command. Start implementation
					of 'info' device.
    10 November 2005 - ver. 0.3.1 (EB): Add event report support for kernel
					buffer size for 'card' write
    12 November 2005 - ver. 0.3.5 (EB): Upgrade to Linux kernel 2.6.14.2.
					Source code improvement.
					Debug. Mean of "pause" command changed.
					unlocked_ioctl instead of ioctl.
					New module parameter "wr_ch_interval".
    21 November 2005 - ver. 0.4.0 (EB): Blocking read in card device.
					New module parameter "rd_ch_interval".
					Poll in card device (only for test).
					Re-write for some init procedure.
					Add support for sysfs.
    22 November 2005 - ver. 0.4.1 (EB): New name for ioctl command. Debug.
					Re-write module init procedure.
					Start to modify all module option, they
					will be local for each device and
					settable by sysfs (not complete).
    22 November 2005 - ver. 0.4.2 (EB): Finish to modify module option for
					write buffer of card device,they now
					are local for each card device and
					settable by sysfs.
    23 November 2005 - ver. 0.4.3 (EB): Code re-organize.
					galil1800_send() rewrite.
    26 November 2005 - ver. 0.4.4 (EB): Finish to modify all module option for
					all device,they now are local for each
					device and settable by sysfs.
					Code re-organize. Change method to
					insert "pause" pseudo-command.
    26 November 2005 - ver. 0.4.5 (EB): Add support for fifo2 (now info device
					work). Export "serial" in device sysfs.
    26 November 2005 - ver. 0.4.7 (EB): Fixed some bugs. Add parameter (also
					in sysfs): [read,write,fifo2]_timeout.
					Add "Unsafe stop of write" command.
					Implemented ioctl for info device.
    29 November 2005 - ver. 0.5.0 (EB): Code re-styling and some function
					rewrited.
    25 February 2006   ver. 0.5.1 (EB): Kernel 2.6.15 support; some code
                                        restyling. License GPL v2.
    02 March    2006   ver. 0.5.2 (EB): Removed card poll file operations and
					other related functions and data.
					Write a complete README file.
					Improve debug print.
					Change timeout use for send, recv and
					read_FIFO2 in galil1800_io.c
					Changed default values for
					some parameters.

*/


#include "galil1800_core.h"
#include "galil1800_io.h"
#include "galil1800_init.h"
#include "galil1800_sysfs.h"
#include "galil1800_info.h"

#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <asm/uaccess.h>


struct pci_device_id galil1800_idtable[] = {
	{  .vendor = PLX_PCI_VENDOR_ID,
	   .device = PLX_PCI_DEVICE_ID,
	   .subvendor = GALIL_SUB_VENDOR,
	   .subdevice = GALIL_SUB_DEVICE_1800,
	   .class = 0,
	   .class_mask = 0,
	   .driver_data = 0 },
	{ 0, }
};

int galil1800_probe(struct pci_dev *pcidev,
			const struct pci_device_id *devid) __init;

void galil1800_remove(struct pci_dev *pcidev) __exit;

struct pci_driver galil1800_driver  = {
	.name = GALIL1800_MODULE_NAME,
/* TODO  This is only for alpha version... */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
	.owner = THIS_MODULE,
#endif
	.id_table = galil1800_idtable,
	.probe = galil1800_probe,
	.remove = __exit_p(galil1800_remove),
	.suspend = NULL,
	.resume = NULL,
	.enable_wake = NULL,
	.shutdown = NULL
};


MODULE_AUTHOR("Universit di Roma Tor Vergata (M. Cesati - E. Betti)");
MODULE_DESCRIPTION("Galil 1800 Motion Controller Driver for kernel 2.6");
MODULE_VERSION("0.5.2");
MODULE_ALIAS("galil1800");
MODULE_LICENSE("GPL v2");
MODULE_DEVICE_TABLE(pci, galil1800_idtable);


unsigned int major = 0; /* major number of device file */

/* the size of the driver's Kernel Mode buffer of interrupt status values
   and others special driver events.
   This is the number of events that can be saved whitout overwrite. */
unsigned int event_buff_sz = GALIL1800_EVENT_BUFF_DEFAULT_SZ;

/* Size for write buffer in card device, set 0 to disable the buffer */
unsigned int card_wbuff_sz = GALIL1800_CARD_WBUFF_DEFAULT_SZ;

/* ticks to wait for check available space in the write fifo.
   Now is initialized in milliseconds and user must set it in ms,
   after, in module init function or ioctl, it is checked and converted! */
unsigned int write_delay = GALIL1800_WRITE_DELAY_MS_DEFAULT;

/* ticks to wait for check available byte in the read fifo.
   Now is initialized in milliseconds and user must set it in ms,
   after, in module init function or ioctl, it is checked and converted! */
unsigned int read_delay = GALIL1800_READ_DELAY_MS_DEFAULT;

/* timeout (in ticks) to wait when there is no space in the write fifo:
   in this time we continuosly check the fifo.
   Now is initialized in milliseconds and user must set it in ms,
   after, in module init function or ioctl, it is checked and converted! */
unsigned int write_timeout = GALIL1800_WRITE_TIMEOUT_MS_DEFAULT;

/* timeout (in ticks) to wait when there is no data in the read fifo:
   in this time we continuosly check the fifo.
   Now is initialized in milliseconds and user must set it in ms,
   after, in module init function or ioctl, it is checked and converted! */
unsigned int read_timeout = GALIL1800_READ_TIMEOUT_MS_DEFAULT;

/* timeout (in ticks) to wait when there is no data in the secondary fifo:
   in this time we continuosly check the fifo.
   Now is initialized in milliseconds and user must set it in ms,
   after, in module init function or ioctl, it is checked and converted! */
unsigned int fifo2_timeout = GALIL1800_FIFO2_TIMEOUT_MS_DEFAULT;

/* "pause" pseudo-command on (1) or off (0) */
unsigned int pause_enable = GALIL1800_CARD_PAUSE_ENABLE_DEFAULT;


module_param(major, uint, S_IRUGO);
module_param(event_buff_sz, uint, S_IRUGO);
module_param(card_wbuff_sz, uint, S_IRUGO);
module_param(write_delay, uint, S_IRUGO);
module_param(read_delay, uint, S_IRUGO);
module_param(write_timeout, uint, S_IRUGO);
module_param(read_timeout, uint, S_IRUGO);
module_param(fifo2_timeout, uint, S_IRUGO);
module_param(pause_enable, uint, S_IRUGO);

MODULE_PARM_DESC(major, "Major number of the device file");
MODULE_PARM_DESC(event_buff_sz, "Buffer size of interrupt "
			"status values and other special driver events");
MODULE_PARM_DESC(card_wbuff_sz, "Size of card write buffer (in bytes),"
			" set 0 to disable");
MODULE_PARM_DESC(write_delay, "Interval to wait for check available"
			" space in the write fifo (in ms)");
MODULE_PARM_DESC(read_delay, "Interval to wait for check available byte"
			" in the read fifo (in ms)");
MODULE_PARM_DESC(write_timeout, "Timeout (in ms) to wait when there is no"
			" space in the write fifo");
MODULE_PARM_DESC(read_timeout, "Timeout (in ms) to wait when there is no"
			" data in the read fifo");
MODULE_PARM_DESC(fifo2_timeout, "Timeout (in ms) to wait when there is no"
			" data in the secondary fifo");
MODULE_PARM_DESC(pause_enable, "\"pause\" pseudo command (by ioctl) enable (1)"
			" or disable (0)");


struct galil1800_module_struct {
	dev_t dev;     /* first major/minor number owned by the driver */
	struct list_head cards; /* list of galil1800_state descriptors */
	int count;               /* counter of registered card devices */
	spinlock_t lock;                        /* protect this struct */
	struct class sysfs_class;                   /* class for sysfs */
} galil1800_module;

/* init for module struct, called in module init function */
#define init_galil1800_module_struct(_gm) do {                          \
	memset((char*)(_gm),0,sizeof(struct galil1800_module_struct));  \
	INIT_LIST_HEAD(&((_gm)->cards));                                \
	(_gm)->count = 0;                                               \
	spin_lock_init(&((_gm)->lock));                                 \
}while(0)


/* Called in probe function */
static void __init init_galil1800_state(struct galil1800_state *s,
		struct pci_dev *pcidev, struct galil1800_module_struct *gm)
{
	INIT_LIST_HEAD(&s->devs);
	s->sysfs_class = &gm->sysfs_class;
	s->dev = pcidev;
	s->io = pci_resource_start(pcidev, GALIL1800_REG_BAR_INDEX);
	s->ctrl_reg = s->io + 4;
	s->irq_reg = s->io + 8;
	s->chan2_reg = s->io + 12;
/* TODO	s->chan2_dpram = pci_resource_start(pcidev,
				GALIL1800_DPRAM_BAR_INDEX); */
	s->irq = pcidev->irq;
	spin_lock_init(&(s->hwlock));
	/* N.B. s->index and s->serial are initialized in probe funcion */

	s->dev_attr_serial.attr.name = "serial";
	s->dev_attr_serial.attr.owner = THIS_MODULE;
	s->dev_attr_serial.attr.mode = S_IRUGO;
	s->dev_attr_serial.show  = galil1800_sysfs_serial_show;
	s->dev_attr_serial.store = NULL;

	s->open_devs = 0;
	init_MUTEX(&(s->open_devs_sem));
	init_GALIL1800_DEVFILE_STATE_array(s->devfile_state);

	init_galil1800_card_wbuffer(&(s->card_wbuff), card_wbuff_sz, pause_enable);

	init_waitqueue_head(&(s->card_read_wait));

	s->write_delay = write_delay;
	s->read_delay = read_delay;
	s->write_timeout = write_timeout;
	s->read_timeout = read_timeout;

	/* N.B.: all class_device_attribute for card device
	         are initialized in galil1800_card_cdev_add() */

	init_galil1800_event_buffer(&(s->event_buff), event_buff_sz);
	/* N.B.: s->irqstack -- Initilized in open step init function */
	tasklet_init(&(s->irq_tasklet),
		galil1800_bottom_half_interrupt,(unsigned long)s);

	s->fifo2_timeout = fifo2_timeout;
	/* N.B.: class_device_attribute for info device
	         are initialized in galil1800_card_cdev_add() */
}

/* "Probe" function: it runs one time for each card detected */
int __init galil1800_probe(struct pci_dev *pcidev,
			const struct pci_device_id *devid)
{
	struct galil1800_state *s;
	int ret;

	if (pcidev->irq == 0) {
		printk(KERN_ERR PFX "IRQ in pci_dev is 0\n");
		return -1;
	}

	if(pci_resource_len(pcidev, GALIL1800_REG_BAR_INDEX) !=
					GALIL1800_REG_PCI_RESOURCE_LEN) {
		printk(KERN_ERR PFX "pci resource len is not %d\n",
					GALIL1800_REG_PCI_RESOURCE_LEN);
		return -1;
	}

	if (!(s = kzalloc(sizeof(struct galil1800_state), GFP_KERNEL))) {
		printk(KERN_ERR PFX "alloc of galil1800_state struct failed\n");
		return -ENOMEM;
	}

	init_galil1800_state(s, pcidev, &galil1800_module);
	pci_set_drvdata(pcidev, s);

	if (!request_region(s->io, GALIL1800_REG_PCI_RESOURCE_LEN,
						GALIL1800_MODULE_NAME)) {
		printk(KERN_ERR PFX "io ports %#lx->%#lx in use\n", s->io,
				s->io + GALIL1800_REG_PCI_RESOURCE_LEN - 1);
		ret = -1;
		goto err_region;
	}

	ret = pci_enable_device(pcidev);
	if (ret) {
		printk(KERN_ERR PFX "cannot enable PCI device\n");
		goto err_pci;
	}

	/* put it into driver list */
	spin_lock(&galil1800_module.lock);
	list_add_tail(&s->devs, &galil1800_module.cards);
	s->index = galil1800_module.count++; /* set card device index */
	spin_unlock(&galil1800_module.lock);

	ret = galil1800_init_controller_probe_step(s);
	if (ret)
		goto err_init;

	printk(KERN_INFO PFX "device #%d  base I/O: 0x%lx  IRQ: %d  "
			"S/N: \"%s\"\n", s->index, s->io, s->irq, s->serial);

	GALIL_PRINT_DEBUG("galil1800_probe: ended\n");
	return 0;

err_init:
	spin_lock(&galil1800_module.lock);
	list_del(&s->devs);
	galil1800_module.count--;
	spin_unlock(&galil1800_module.lock);
	pci_disable_device(pcidev);
err_pci:
	release_region(s->io, GALIL1800_REG_PCI_RESOURCE_LEN);
err_region:
	kfree(s);
	pci_set_drvdata(pcidev, NULL);
	return ret;
}

/* "Remove" function: it runs one time for each card detected */
void __exit galil1800_remove(struct pci_dev *pcidev)
{
	struct galil1800_state *s = pci_get_drvdata(pcidev);
	if (!s) {
		GALIL_PRINT_DEBUG("galil1800_remove: pci_get_drvdata failed\n");
		return;
	}
	galil1800_clean_controller_remove_step(s);

	spin_lock(&galil1800_module.lock);
	list_del(&s->devs);
	galil1800_module.count--;
	spin_unlock(&galil1800_module.lock);

	pci_disable_device(pcidev);
	release_region(s->io, GALIL1800_REG_PCI_RESOURCE_LEN);
	kfree(s);
	pci_set_drvdata(pcidev, NULL);
	GALIL_PRINT_DEBUG("galil1800_remove: ended\n");
}

int __init galil1800_add_cdev_for_all_cards(void)
{
	int rc, minor=0, card=0;
	struct list_head *s_list;

	list_for_each(s_list, &galil1800_module.cards) {
		struct galil1800_state *s =
			list_entry(s_list, struct galil1800_state, devs);

		rc = galil1800_card_cdev_add(
			&s->devfile_state[GALIL1800_CARD_DEVICE].cdev,
							major, minor);
		if(rc)
			goto out_cdev;
		++minor;

		rc = galil1800_event_cdev_add(
			&s->devfile_state[GALIL1800_EVENT_DEVICE].cdev,
							major, minor);
		if(rc)
			goto out_cdev;
		++minor;

		rc = galil1800_info_cdev_add(
			&s->devfile_state[GALIL1800_INFO_DEVICE].cdev,
							major, minor);
		if(rc)
			goto out_cdev;
		++minor;

		rc = device_create_file(&s->dev->dev, &s->dev_attr_serial);
		if(rc)
			goto out_cdev;
		++card;
	}

	BUG_ON(minor != (GALIL1800_MINOR_PER_CARD * galil1800_module.count));

	return 0;

out_cdev:
	if(!minor)
		return rc;
	list_for_each(s_list, &galil1800_module.cards) {
		struct galil1800_state *s =
			list_entry(s_list, struct galil1800_state, devs);
		if(card) {
			device_remove_file(&s->dev->dev, &s->dev_attr_serial);
			--card;
		}
		galil1800_card_cdev_del(
				&s->devfile_state[GALIL1800_CARD_DEVICE].cdev);
		if(!--minor)
			return rc;
		galil1800_event_cdev_del(
				&s->devfile_state[GALIL1800_EVENT_DEVICE].cdev);
		if(!--minor)
			return rc;
		galil1800_info_cdev_del(
				&s->devfile_state[GALIL1800_INFO_DEVICE].cdev);
		if(!--minor)
			return rc;
	}
	BUG_ON(card!=0 || minor!=0);
	return rc;
}


/* init module function */
int __init galil1800_init(void)
{
	int rc;

	GALIL_PRINT_DEBUG("additional debug print enabled\n");

	galil1800_event_buff_size_check(&card_wbuff_sz);
	galil1800_card_parameter_check(&pause_enable, &card_wbuff_sz);
	galil1800_write_delay_check_and_convert(&write_delay);
	galil1800_read_delay_check_and_convert(&read_delay);
	galil1800_write_timeout_check_and_convert(&write_timeout);
	galil1800_read_timeout_check_and_convert(&read_timeout);
	galil1800_fifo2_timeout_check_and_convert(&fifo2_timeout);

	init_galil1800_module_struct(&galil1800_module);

	/* Following function use galil1800_probe():
	   It runs galil1800_probe() for each card detected */
	rc = pci_register_driver(&galil1800_driver);
	if (rc) {
		printk(KERN_ERR PFX "cannot register the PCI device driver"
				" or no Galil 1800 hardware device found\n");
		return rc;
	}

	if (!galil1800_module.count) {
		printk(KERN_ERR PFX "no Galil 1800 device found\n");
		rc = -ENODEV;
		goto out_dev;
	}

	if (!major) {
		/* Total number of devices = minor_per_card * card_number */
		rc = alloc_chrdev_region(&(galil1800_module.dev), 0,
			GALIL1800_MINOR_PER_CARD * galil1800_module.count,
			GALIL1800_MODULE_NAME);
		major = MAJOR(galil1800_module.dev);
	} else {
		galil1800_module.dev = MKDEV(major, 0);
		rc = register_chrdev_region(galil1800_module.dev,
			GALIL1800_MINOR_PER_CARD * galil1800_module.count,
			GALIL1800_MODULE_NAME);
	}
	if (rc < 0) {
		printk(KERN_ERR PFX "cannot register the device files\n");
		rc = -ENOMEM;
		goto out_dev;
	}

	rc = galil1800_sysfs_class_register(&galil1800_module.sysfs_class);
	if(rc) {
		printk(KERN_ERR PFX "galil1800_sysfs_class_register failed\n");
		goto out_class;
	}

	rc = galil1800_add_cdev_for_all_cards();
	if(rc) {
		printk(KERN_ERR PFX
			"galil1800_add_cdev_for_all_cards failed\n");
		goto out_cdev;
	}

	GALIL_PRINT_DEBUG("module loaded\n");
	return 0;

out_cdev:
	class_unregister(&galil1800_module.sysfs_class);
out_class:
	unregister_chrdev_region(galil1800_module.dev,
		GALIL1800_MINOR_PER_CARD * galil1800_module.count);
out_dev:
	pci_unregister_driver(&galil1800_driver);
	return rc;
}

void __exit galil1800_release_cdevs_for_all_card(void)
{
	int check=0;
	struct list_head *list;
	struct galil1800_state *s;

	list_for_each(list, &galil1800_module.cards) {
		s = list_entry(list, struct galil1800_state, devs);
 		device_remove_file(&s->dev->dev, &s->dev_attr_serial);
		galil1800_card_cdev_del(
			&s->devfile_state[GALIL1800_CARD_DEVICE].cdev);
		galil1800_event_cdev_del(
			&s->devfile_state[GALIL1800_EVENT_DEVICE].cdev);
		galil1800_info_cdev_del(
			&s->devfile_state[GALIL1800_INFO_DEVICE].cdev);
		++check;
	}
	if(check!=galil1800_module.count)
		printk(KERN_ERR PFX "Not all cdevs released\n");
}

/* cleanup module function */
void __exit galil1800_cleanup(void)
{
	galil1800_release_cdevs_for_all_card();

	unregister_chrdev_region(galil1800_module.dev,
		GALIL1800_MINOR_PER_CARD * galil1800_module.count);

	class_unregister(&galil1800_module.sysfs_class);

	/* Following function use galil1800_remove():
	   It runs galil1800_remove() for each card detected */
	pci_unregister_driver(&galil1800_driver);
	/* Now all struct galil1800_state are freed! */

	GALIL_PRINT_DEBUG("module unloaded\n");
}

module_init(galil1800_init);
module_exit(galil1800_cleanup);
